package eu.clarin.cmdi.oai.harvester.impl;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.concurrent.atomic.AtomicReference;

import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

import eu.clarin.cmdi.oai.harvester.HarvestHandler;
import eu.clarin.cmdi.oai.harvester.HarvestJob;
import eu.clarin.cmdi.oai.harvester.HarvesterException;
import eu.clarin.cmdi.oai.harvester.Header;
import eu.clarin.cmdi.oai.harvester.MetadataFormat;
import eu.clarin.cmdi.oai.harvester.Repository;
import eu.clarin.cmdi.oai.harvester.Repository.DeletedNotion;
import eu.clarin.cmdi.oai.harvester.Repository.Granularity;

final class HarvestJobImpl implements HarvestJob {
    private enum Task {
        IDENTIFY_REPOSITORY,
        ENUMERATE_METADATA_FORMATS,
        INITALIZE_HARVEST_RECORDS,
        HARVEST_RECORDS
    }
    private final AbstractHarvester harvester;
    private final String id;
    private final URI repositoryURI;
    private State state = State.NEW;
    private AtomicReference<HarvestHandler> handler;
    private Date from;
    private Date until;
    private String set;
    private List<String> metadataPrefixes;
    private String repositoryName;
    private String baseURL;
    private String protocolVersion;
    private List<String> adminEmail = Collections.emptyList();
    private String earliestTimestamp;
    private DeletedNotion deletedNotion;
    private Granularity granularity;
    private int compressionMask;
    private List<MetadataFormat> metadataFormats;
    private long recordCount = 0;
    private long requestCount = 0;
    private long resumptionCount = 0;
    private long bytedTransferred = 0;
    private long timeTotal = 0;
    private long timeNetwork = 0;
    private long timeWait = 0;
    private long timeProcessing = 0;
    private long startTime;
    private Task task = Task.IDENTIFY_REPOSITORY;
    private List<String> availablePrefixes;
    private String resumptionToken;
    private Date latestDatestamp;

    HarvestJobImpl(AbstractHarvester harvester, String id, URI repositoryURI,
            HarvestHandler handler) {
        this.harvester = harvester;
        this.id = id;
        this.repositoryURI = repositoryURI;
        this.handler = new AtomicReference<HarvestHandler>(handler);
    }

    @Override
    public String getId() {
        return id;
    }

    @Override
    public State getState() {
        return state;
    }

    @Override
    public long getRecordCount() {
        return recordCount;
    }

    @Override
    public long getRequestCount() {
        return requestCount;
    }

    @Override
    public long getResumptionCount() {
        return resumptionCount;
    }

    @Override
    public long getBytesTransferred() {
        return bytedTransferred;
    }

    @Override
    public long getTotalTime() {
        return timeTotal;
    }

    @Override
    public long getNetworkTime() {
        return timeNetwork;
    }
    
    @Override
    public long getWaitTime() {
        return timeWait;
    }

    @Override
    public long getProcessingTime() {
        return timeProcessing;
    }

    
    @Override
    public HarvestHandler getHarvestHandler() {
        return handler.get();
    }
    
    @Override
    public void setHarvestHandler(HarvestHandler handler) {
        this.handler.set(handler);
    }

    @Override
    public Date getFrom() {
        return from;
    }
    
    @Override
    public void setFrom(Date from) {
        if (state != State.NEW) {
            throw new IllegalStateException("state != new");
        }
        this.from = from;
    }
    
    @Override
    public Date getUntil() {
        return until;
    }
    
    @Override
    public void setUntil(Date until) {
        if (state != State.NEW) {
            throw new IllegalStateException("state != new");
        }
        this.until = until;
    }

    @Override
    public String getSet() {
        return set;
    }

    @Override
    public void setSet(String set) {
        if (state != State.NEW) {
            throw new IllegalStateException("state != new");
        }
        if ((set != null) && !set.isEmpty()) {
            this.set = set;
        } else {
            this.set = null;
        }
    }

    @Override
    public List<String> getMetadataPrefixes() {
        return metadataPrefixes;
    }
    
    @Override
    public void setMetadataPrefixes(List<String> metadataPrefixes) {
        if (state != State.NEW) {
            throw new IllegalStateException("state != new");
        }
        if ((metadataPrefixes != null) && !metadataPrefixes.isEmpty()) {
            this.metadataPrefixes = metadataPrefixes;
        } else {
            this.metadataPrefixes = null;
        }
    }

    @Override
    public void run() throws HarvesterException {
        harvester.doRunJob(this);
    }

    @Override
    public void cancel() throws HarvesterException {
        state = State.CANCELED;
        harvester.doCancelJob(this);
    }

    public URI getRepositoryURI() {
        return repositoryURI;
    }

    void setState(State state) {
        this.state = state;
    }

    void setRepositoryName(String repositoryName) {
        this.repositoryName = repositoryName;
    }

    void setBaseURL(String baseURL) {
        this.baseURL = baseURL;
    }

    void setProtocolVersion(String protocolVersion) {
        this.protocolVersion = protocolVersion;
    }
    
    void setAdminEmail(List<String> adminEmail) {
        this.adminEmail = adminEmail;
    }

    void setEarliestTimestamp(String earliestTimestamp) {
        this.earliestTimestamp = earliestTimestamp;
    }

    void setDeletedNotion(DeletedNotion deletedNotion) {
        this.deletedNotion = deletedNotion;
    }

    void setGranularity(Granularity granularity) {
        this.granularity = granularity;
    }

    Granularity getGranularity() {
        return granularity;
    }

    void setCompressionMask(int compressionMask) {
        this.compressionMask = compressionMask;
    }
    
    int getCompressionMask() {
        return compressionMask;
    }

    void incRequestCount() {
        requestCount++;
    }

    void incResumptionCount() {
        resumptionCount++;
    }

    void addToBytesTransferred(long bytes) {
        bytedTransferred += bytes;
    }

    void addToNetworkTime(long delta) {
        timeNetwork += delta;
    }

    void addToWaitTime(long delta) {
        timeWait += delta;
    }


    void addToProcessingTime(long delta) {
       timeProcessing += delta;
    }

    String getResumptionToken() {
        return resumptionToken;
    }
    
    void setResumptionToken(String resumptionToken) {
        this.resumptionToken = resumptionToken;
    }

    String getCurrentPrefix() {
        return availablePrefixes.get(0);
    }

    boolean process(HarvestWorker worker) throws HarvesterException {
        if (state == State.CANCELED) {
            return true;
        }
        if (state != State.RUNNING) {
            throw new HarvesterException("job is not in running state");
        }
        try {
            switch (task) {
            case IDENTIFY_REPOSITORY:
                startTime = System.currentTimeMillis();
                worker.doIdentify(this);
                task = Task.ENUMERATE_METADATA_FORMATS;
                break;
            case ENUMERATE_METADATA_FORMATS:
                worker.doListMetadataFormats(this);
                task = Task.INITALIZE_HARVEST_RECORDS;
                break;
            case INITALIZE_HARVEST_RECORDS:
                if (availablePrefixes == null) {
                    availablePrefixes = new ArrayList<String>();
                    if (metadataPrefixes != null) {
                        availablePrefixes.addAll(metadataPrefixes);
                    } else if (metadataFormats != null) {
                        for (MetadataFormat format : metadataFormats) {
                            availablePrefixes.add(format.getPrefix());
                        }
                    } else {
                        availablePrefixes.add("oai_dc");
                    }
                }
                task = Task.HARVEST_RECORDS;
                resumptionToken = null;
                latestDatestamp = null;
                onStartListingRecords(getCurrentPrefix());
                /* FALL-TROUGH */
            case HARVEST_RECORDS:
                boolean more = worker.doListRecords(this);
                if (!more) {
                    onFinishListingRecords(getCurrentPrefix());
                    availablePrefixes.remove(0);
                    if (availablePrefixes.isEmpty()) {
                        timeTotal = System.currentTimeMillis() - startTime;
                        state = State.DONE;
                        return true;
                    } else {
                        task = Task.INITALIZE_HARVEST_RECORDS;
                    }
                }
            }
        } catch (Throwable e) {
            timeTotal = System.currentTimeMillis() - startTime;
            state = State.ERROR;
            if (e instanceof HarvesterException) {
                throw (HarvesterException) e;
            } else if (e instanceof XMLStreamException) {
                throw new HarvesterException("malformed OAI response", e);
            } else if (e instanceof IOException) {
                throw new HarvesterException("network error", e);
            } else {
                throw new HarvesterException("unexpected exception", e);
            }
        }
        return false;
    }

    void onStartListingRecords(String prefix) {
        final HarvestHandler h = handler.get();
        if (h != null) {
            h.onStartListingRecords(prefix);
        }
    }

    void onFinishListingRecords(String prefix) {
        final HarvestHandler h = handler.get();
        if (h != null) {
            h.onFinishListingRecords(prefix, latestDatestamp);
        }
    }

    void onRecord(Header header) {
        recordCount++;
        if (latestDatestamp == null) {
            latestDatestamp = header.getDatestamp();
        } else {
            if (latestDatestamp.compareTo(header.getDatestamp()) < 0) {
                latestDatestamp = header.getDatestamp();
            }
        }
        final HarvestHandler h = handler.get();
        if (h != null) {
            h.onRecord(header);
        }
    }

    void onRecordMetadata(Header header, XMLStreamReader reader) {
        final HarvestHandler h = handler.get();
        if (h != null) {
            h.onRecordMetadata(header, reader);
        }
    }

    void onRecordAbout(Header header, XMLStreamReader reader) {
        final HarvestHandler h = handler.get();
        if (h != null) {
            h.onRecordAbout(header, reader);
        }
    }

    void onIdentify() {
        final HarvestHandler h = handler.get();
        if (h != null) {
            final Repository repository = new Repository() {
                @Override
                public String getRepositoryName() {
                    return repositoryName;
                }
                
                @Override
                public String getProtocolVersion() {
                    return protocolVersion;
                }
                
                @Override
                public Granularity getGranularity() {
                    return granularity;
                }
                
                @Override
                public String getEarliestTimestamp() {
                    return earliestTimestamp;
                }
                
                @Override
                public DeletedNotion getDeletedNotion() {
                    return deletedNotion;
                }
                
                @Override
                public int getCompressionMask() {
                    return compressionMask;
                }
                
                @Override
                public String getBaseURL() {
                    return baseURL;
                }
                
                @Override
                public List<String> getAdminEmail() {
                    return adminEmail;
                }
            };
            h.onIdentify(repository);
        }
    }

    void onListMetadataFormats(List<MetadataFormat> metadataFormats) {
        this.metadataFormats = metadataFormats;
        final HarvestHandler h = handler.get();
        if (h != null) {
            h.onListMetadataFormats(this.metadataFormats);
        }
    }

    InputStream wrap(InputStream stream) throws IOException {
        final HarvestHandler h = handler.get();
        if (h != null) {
            return h.wrap(stream);
        } else {
            return stream;
        }
    }

} // class HarvestJob
