package eu.clarin.cmdi.oai.harvester.impl;

import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.zip.GZIPInputStream;
import java.util.zip.Inflater;
import java.util.zip.InflaterInputStream;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

import org.apache.http.HttpEntity;

import eu.clarin.cmdi.oai.harvester.HarvesterException;
import eu.clarin.cmdi.oai.harvester.HarvesterProtocolErrorException;
import eu.clarin.cmdi.oai.harvester.ProtocolError;

final class Response {
    private final class CountingInputStream extends FilterInputStream {
        private long count = 0;

        private CountingInputStream(final InputStream stream) {
            super(stream);
        }

        @Override
        public int read() throws IOException {
            final int result = super.read();
            count += (result > 0) ? 1 : 0;
            return result;
        }

        @Override
        public int read(byte[] buffer, int offset, int length)
                throws IOException {
            final int result = super.read(buffer, offset, length);
            count += (result > 0) ? result : 0;
            return result;
        }

        @Override
        public int read(byte[] buffer) throws IOException {
            final int result = super.read(buffer);
            count += (result > 0) ? result : 0;
            return result;
        }

        @Override
        public long skip(long n) throws IOException {
            final long result = super.skip(n);
            count += (result > 0) ? n : 0;
            return result;
        }
        
        public long getByteCount() {
            return count;
        }

    } // class CountingInputStream

    private static final String OAI_NS =
        "http://www.openarchives.org/OAI/2.0/";
    private final HarvestJobImpl job;
    private final HttpEntity entity;
    private final long now = System.currentTimeMillis();
    private CountingInputStream stream; 
    private XMLStreamReader reader;

    Response(HarvestJobImpl job, HttpEntity entity, AbstractHarvester harvester)
            throws IOException, XMLStreamException, HarvesterException {
        this.job = job;
        this.entity = entity;

        if (entity.getContentType() != null) {
            System.err.println("Content-Type: " +
                    entity.getContentType().getValue());
            System.err.println("Content-Length: " + entity.getContentLength());
        }
        InputStream in = entity.getContent();
        org.apache.http.Header enc = entity.getContentEncoding();
        if (enc != null) {
            System.err.println("Content-Encoding: " +
                    entity.getContentEncoding().getValue());
            if ("gzip".equalsIgnoreCase(enc.getValue())) {
                in = new GZIPInputStream(in);
            } else if ("deflate".equalsIgnoreCase(enc.getValue())) {
                in = new InflaterInputStream(in, new Inflater(true));
            }
        }
        this.stream = new CountingInputStream(job.wrap(in));
        this.reader = harvester.createReader(stream);
        readStart(OAI_NS, "OAI-PMH", true);
        readContent(OAI_NS, "responseDate", true);
        readStart(OAI_NS, "request", true, true);
        // System.err.println("YYY XXX = " + toReadable(reader) + ", " +
        // reader.getAttributeValue(null, "verb"));
        readEnd(OAI_NS, "request", true);

        List<ProtocolError> errors = null;
        while (readStart(OAI_NS, "error", false, true)) {
            if (errors == null) {
                errors = new ArrayList<ProtocolError>();
            }
            final String code = reader.getAttributeValue(null, "code");
            reader.next(); // skip start element after reading attrs
            final String msg = readString(true);
            readEnd(OAI_NS, "error");
            errors.add(new ProtocolError(code, msg));
        }
        if (errors != null) {
            throw new HarvesterProtocolErrorException(errors);
        }
    }
    
    public void close() throws XMLStreamException {
        readEnd(OAI_NS, "OAI-PMH");
    }

    public void release() {
        job.addToProcessingTime(System.currentTimeMillis() - now);
        job.addToBytesTransferred(stream.getByteCount());
        if (reader != null) {
            try {
                reader.close();
            } catch (XMLStreamException e) {
                /* IGNORE */
            }
        }
        if (stream != null) {
            try {
                stream.close();
            } catch (IOException e) {
                /* IGNORE */
            }
        }
        try {
            entity.consumeContent();
        } catch (IOException e) {
            /* IGNORE */
        }
    }

    public boolean readStart(String namespaceURI, String localName,
            boolean required) throws XMLStreamException {
        return readStart(namespaceURI, localName, required, false);
    }

    public boolean readStart(String namespaceURI, String localName,
            boolean required, boolean attributes) throws XMLStreamException {
//        System.err.println("readStart (" + localName + ", required = " +
//                required + ") @ " + toReadable(reader));
        if (!reader.isEndElement()) {
            while (reader.hasNext()) {
//                System.err.println("  LOOP: " + dumpState());
                if (reader.isWhiteSpace()) {
                    reader.next();
                    continue;
                }
                if (reader.isStartElement()) {
                    if (namespaceURI.equals(reader.getNamespaceURI()) &&
                            localName.equals(reader.getLocalName())) {
//                        System.err.print("--> found ");
                        if (!attributes) {
//                            System.err.print("and consumed ");
                            reader.next(); // skip to next event
                        }
//                        System.err.println("@ " + toReadable(reader));
                        return true;
                    }
                    break;
                }
                if (reader.isCharacters() || reader.isEndElement()) {
                    break;
                }
                reader.next();
            } // while
        }
        if (required) {
//            System.err.println("--> error, not found @ " + toReadable(reader));
            throw new XMLStreamException("expected element '" +
                    new QName(namespaceURI, localName) + "', but found '" +
                    reader.getName() + "'", reader.getLocation());
        }
//        System.err.println("--> not found @ " + toReadable(reader));
        return false;
    }

    public void readEnd(String namespaceURI, String localName)
            throws XMLStreamException {
        readEnd(namespaceURI, localName, false);
    }

    public void readEnd(String namespaceURI, String localName,
            boolean skipContent) throws XMLStreamException {
//        System.err.println("readEnd (" + localName + ") @ " + dumpState() +
//                ", skipContent = " + skipContent);
        int level = 1;
        while (reader.hasNext()) {
//            System.err.println("  LOOP " + dumpState() + " [" +
//                    level + "]");
            if (reader.isWhiteSpace()) {
                reader.next();
                continue;
            }
            if (skipContent) {
                if (reader.isCharacters()) {
                    reader.next();
                    continue;
                }
                if (reader.isStartElement()) {
                    if (!(namespaceURI.equals(reader.getNamespaceURI()) &&
                            localName.equals(reader.getLocalName()))) {
                        level++;
                    }
                    reader.next();
                    continue;
                }
            }
            if (reader.isEndElement()) {
                level--;
//                System.err.println("   @END-TAG: " + dumpState() + " [" +
//                        level + "]");
                if (level == 0) {
                    if (namespaceURI.equals(reader.getNamespaceURI()) &&
                        localName.equals(reader.getLocalName())) {
                        reader.next(); // consume tag
                        break;
                    } else {
                        throw new XMLStreamException("expected end tag for '" +
                                new QName(namespaceURI, localName) +
                                "', but found '" + reader.getName() + "'",
                                reader.getLocation());
                    }
                }
            }
            reader.next();
        }
//        System.err.println("--> ok @ " + dumpState());
    }

    public String readContent(String namespaceURI, String localName,
            boolean required) throws XMLStreamException {
        String result = null;
        if (readStart(namespaceURI, localName, required)) {
            result = readString(true);
            readEnd(namespaceURI, localName);
        }
        return result;
    }

    public String readString(boolean required)
            throws XMLStreamException {
//        System.err.println("readString @ " + toReadable(reader));
        String s = null;
        if (reader.isCharacters()) {
            s = reader.getText();
            if (s != null) {
                s = s.trim();
            }
            reader.next();
        }
        if (required && ((s == null) || s.isEmpty())) {
            throw new XMLStreamException("expected character content "
                    + "at position", reader.getLocation());
        }
//      System.err.println("--> ok @ " + toReadable(reader));
        return s;
    }

    public String readAttributeValue(String namespaceURI, String localName)
            throws XMLStreamException {
        if (!reader.isStartElement()) {
            throw new XMLStreamException("not at a start elment event",
                    reader.getLocation());
        }
        String attr = reader.getAttributeValue(namespaceURI, localName);
        if (attr != null) {
            attr = attr.trim().intern();
        }
        return attr;
    }

    public String readNamespaceURI() throws XMLStreamException {
        if (!reader.isStartElement()) {
            throw new XMLStreamException("not at a start elment event",
                    reader.getLocation());
        }
        return reader.getNamespaceURI();
    }

    public void consumeStart() throws XMLStreamException {
        if (!reader.isStartElement()) {
            throw new XMLStreamException("not at a start elment event",
                    reader.getLocation());
        }
        reader.next();
    }

    public void consumeWhitespace() throws XMLStreamException {
        while (reader.isWhiteSpace() && reader.hasNext()) {
            reader.next();
            continue;
        }
    }

    public XMLStreamReader getXMLStreamReader() {
        return reader;
    }

    public String dumpState() {
        StringBuilder sb = new StringBuilder();
        switch (reader.getEventType()) {
        case XMLStreamConstants.START_DOCUMENT:
            return "START_DOC";
        case XMLStreamConstants.END_DOCUMENT:
            return "END_DOC";
        case XMLStreamConstants.START_ELEMENT:
            sb.append("START[");
            sb.append(reader.getNamespaceURI());
            sb.append(",");
            sb.append(reader.getLocalName());
            sb.append("]");
            break;
        case XMLStreamConstants.END_ELEMENT:
            sb.append("END[");
            sb.append(reader.getNamespaceURI());
            sb.append(",");
            sb.append(reader.getLocalName());
            sb.append("]");
            break;
        case XMLStreamConstants.CHARACTERS:
            sb.append("CHARACTERS[\"");
            sb.append(reader.getText().replace("\n", "\\n")
                    .replace("\r", "\\r").replace("\t", "\\t"));
            sb.append("\", isWhitespace = ");
            sb.append(reader.isWhiteSpace());
            sb.append("]");
            break;
        case XMLStreamConstants.CDATA:
            sb.append("CDATA[\"");
            sb.append(reader.getText().replace("\n", "\\n")
                    .replace("\r", "\\r").replace("\t", "\\t"));
            sb.append("\", isWhitespace = ");
            sb.append(reader.isWhiteSpace());
            sb.append("]");
            break;
        default:
            sb.append(Integer.toString(reader.getEventType()));
        }
        return sb.toString();
    }
    
} // class XmlStreamReaderWrapper
